#ifndef __grid__
   #define __grid__
   
   #include once "core.bi"
   #include once "vec2h.bi"
   #include once "mat3.bi"
   
   /'
      A bounding box (or bounding rectangle) object
      
      this is useful for clipping of objects, geometry, and collision detection, to
      name a few applications.
      The grid object derives from here, and also provides spatial partition functionality
      besides the bounding functionality
   '/
   class boundingBox extends object
      protected:
         as mat3 m_transform = any
         as mat3 m_invTransform = any
         
         as double m_width = any
         as double m_height = any
         
      public:
         declare constructor()
         declare constructor( byval w as double, byval h as double )
               
         declare property transform() as mat3
         declare property transform( byval M as mat3 )
         
         declare property invTransform() as mat3
         
         declare virtual property width() as double
         declare virtual property width( byval value as double )
         
         declare virtual property height() as double
         declare virtual property height( byval value as double )
         
         declare function isInside( byval p as vec2h ) as boolean
         declare function getPoint( byval p as vec2h ) as vec2h
         
         declare virtual sub render( byval c as uint32 = rgba( 222, 222, 222, 255 ) )
   end class

   constructor boundingBox()
      m_width = 100.0
      m_height = 100.0
      
      m_transform = matrices.identity2h()
      m_invTransform = inverse( m_transform )
   end constructor
   
   constructor boundingBox( byval w as double, byval h as double )
      m_width = iif( w > 0, w, 100.0 )
      m_height = iif( h > 0, h, 100.0 )
      
      m_transform = matrices.identity2h()
      m_invTransform = inverse( m_transform )
      
   end constructor
   
   property boundingBox.transform() as mat3
      return( m_transform )
   end property
   
   property boundingBox.transform( byval M as mat3 )
      m_transform = M
      '' update also the inverse matrix of the transformation
      m_invTransform = inverse( m_transform )
   end property
   
   property boundingBox.invTransform() as mat3
      return( m_invTransform )
   end property
   
   property boundingBox.width() as double
      return( m_width )
   end property
   
   property boundingBox.width( byval value as double )
      m_width = value
   end property
   
   property boundingBox.height() as double
      return( m_height )
   end property
   
   property boundingBox.height( byval value as double )
      m_height = value
   end property
   
   function boundingBox.getPoint( byval p as vec2h ) as vec2h
      '' returns the point in bounding box coordinates
      return( m_invTransform * p )
   end function
   
   function boundingBox.isInside( byval p as vec2h ) as boolean
      /'
         this function returns true if a point p lies inside the bounding box
              
         0,0 +-------+
             |       |
             |   p   |
             |       |
             +-------+
                       m_width, m_height
         
         it works transforming the point (which must be specified in global coordinates)
         to the coordinate system of the bounding box, by multiplying the point with the
         inverse transformation matrix of the bounding box.
         After the transformation, the test is straightforward.
      '/
      dim as vec2h tp = getPoint( p )
      
      return( iif( tp.x >= 0 andAlso tp.x <= m_width - 1 andAlso tp.y >= 0 andAlso tp.y <= m_height - 1, true, false ) )
   end function
   
   sub boundingBox.render( byval c as uint32 = rgba( 222, 222, 222, 255 ) )
      '' renders the bounding box (mostly for debug purposes)
      
      /'
            p0 +-----+ p3
               |     |
               |     |
               |     |
            p1 +-----+ p2
      '/
      '' transform the origin and derive the rest of the points from there         
      dim as vec2h p0 = m_transform * vec2h( 0.0, 0.0 )
      dim as vec2h p1 = m_transform * vec2h( 0.0, m_height )
      dim as vec2h p2 = m_transform * vec2h( m_width, m_height )
      dim as vec2h p3 = m_transform * vec2h( m_width, 0.0 )
            
      line( p0.x, p0.y ) - ( p1.x, p1.y ), c
      line( p1.x, p1.y ) - ( p2.x, p2.y ), c
      line( p2.x, p2.y ) - ( p3.x, p3.y ), c
      line( p3.x, p3.y ) - ( p0.x, p0.y ), c
      
      dim as vec2h center = vec2h( p0.x + m_width / 2, p0.y + m_height / 2 )
      
      circle( center.x, center.y ), 6, rgba( 255, 0, 255, 255 ), , , , f
   end sub

   /'
      grid object
      
      A grid object that can be used as an acceleration structure
      Derives from boundingBox
   '/
   class grid extends boundingBox
      protected:
         as double m_cellWidth
         as double m_cellHeight
         as integer m_gridXSize
         as integer m_gridYSize
      
      public:
         declare constructor()
         declare constructor( byval w as double, byval h as double, byval xSize as integer, byval ySize as integer )
         declare destructor()
   
         declare property width() as double
         declare property width( byval value as double )
         
         declare property height() as double
         declare property height( byval value as double )
         
         declare property gridXSize() as integer
         declare property gridXSize( byval value as integer )
         
         declare property gridYSize() as integer
         declare property gridYSize( byval value as integer )
         
         declare property cellWidth() as double
         declare property cellHeight() as double
   
         declare function getCell( byval p as vec2h ) as vec2h
         
         declare sub render( byval c as uint32 = rgba( 200, 200, 200, 255 ) )         
   end class
      
   constructor grid( byval w as double, byval h as double, byval xSize as integer, byval ySize as integer )      
      '' call the base constructor to initialize
      base.constructor( w, h )
         
      m_gridXSize = iif( xSize > 0, xSize, 1 )
      m_gridYSize = iif( ySize > 0, ySize, 1 )      
      m_cellWidth = m_width / m_gridXSize
      m_cellHeight = m_height / m_gridYSize
   end constructor
   
   destructor grid()
   
   end destructor
   
   property grid.cellWidth() as double
      return( m_cellWidth )
   end property
   
   property grid.cellHeight() as double
      return( m_cellHeight )
   end property
   
   property grid.width() as double
      return( m_width )
   end property
   
   property grid.width( byval value as double )
      m_width = value
      m_cellWidth = m_width / m_gridXSize   
   end property
   
   property grid.height() as double
      return( m_height )
   end property
   
   property grid.height( byval value as double )
      m_height = value
      m_cellHeight = m_height / m_gridYSize
   end property
   
   property grid.gridXSize() as integer
      return( m_gridXSize )
   end property
   
   property grid.gridXSize( byval value as integer )
      m_gridXSize = value
      m_cellWidth = m_width / m_gridXSize   
   end property
   
   property grid.gridYSize() as integer
      return( m_gridYSize )
   end property
   
   property grid.gridYSize( byval value as integer )
      m_gridYSize = value
      m_cellHeight = m_height / m_gridYSize
   end property
   
   function grid.getCell( byval p as vec2h ) as vec2h
      '' returns the cell in the grid this point lies on
      '' assumes the point is already in grid space coordinates
      return( vec2h( p.x / m_cellWidth, p.y / m_cellHeight ) )   
   end function

   sub grid.render( byval c as uint32 = rgba( 200, 200, 200, 255 ) )
      /'
         get the 4 corners of the grid
         
         p0-----p3
         |       |
         |       |
         p1-----p2
      '/      
      '' transform the grid corners to grid space
      dim as vec2h p0 = m_transform * vec2h( 0.0, 0.0 )
      dim as vec2h p1 = m_transform * vec2h( 0.0, m_height )
      dim as vec2h p2 = m_transform * vec2h( m_width, m_height )
      dim as vec2h p3 = m_transform * vec2h( m_width, 0.0 )
      
      '' draw the borders
      line( p0.x, p0.y ) - ( p1.x, p1.y ), c
      line( p1.x, p1.y ) - ( p2.x, p2.y ), c
      line( p2.x, p2.y ) - ( p3.x, p3.y ), c
      line( p3.x, p3.y ) - ( p0.x, p0.y ), c
      
      '' horizontal lines
      for y as double = 0 to 1 step 1 / m_gridYSize
         dim as vec2h st = interpolate( p0, p1, y )
         dim as vec2h en = interpolate( p3, p2, y )
         
         line( st.x, st.y ) - ( en.x, en.y ), c
      next
      
      '' vertical lines
      for x as double = 0 to 1 step 1 / m_gridXSize
         dim as vec2h st = interpolate( p0, p3, x )
         dim as vec2h en = interpolate( p1, p2, x )
         
         line( st.x, st.y ) - ( en.x, en.y ), c
      next      
   end sub
#endif
