#ifndef __shape__
   #define __shape__
   /'
      shapeEdge type definition
      
      this type simply holds two vertices that define an edge of the shape
      fairly straightforward
   '/   
   #include once "core.bi"
   #include once "vec2h.bi"
   #include once "mat3.bi"
   #include once "arrayList.bi"
   
   type shapeEdge
      public:
         as vec2h ptr v1 = any
         as vec2h ptr v2 = any
      
      declare constructor()
      declare constructor( byval p1 as vec2h ptr, byval p2 as vec2h ptr )
      declare constructor( byref rhs as shapeEdge )
      
      declare operator let( byref rhs as shapeEdge )
   end type
   
   constructor shapeEdge()
      v1 = 0
      v2 = 0
   end constructor
   
   constructor shapeEdge( byval p1 as vec2h ptr, byval p2 as vec2h ptr )
      v1 = p1
      v2 = p2
   end constructor
   
   constructor shapeEdge( byref e as shapeEdge )
      v1 = e.v1
      v2 = e.v2
   end constructor
   
   operator shapeEdge.let( byref rhs as shapeEdge )
      v1 = rhs.v1
      v2 = rhs.v2
   end operator
   
   /'
      shape class
      
      responsibility: this class represents a shape, defined by a collection of vertices and
         the edges that connect them
      mutability: mutable
   '/
   class shape
      public:
         declare constructor()
         declare destructor()
         
         declare property vertexCount() as integer
         declare property edgeCount() as integer
         
         declare function vertex( byval index as integer ) as vec2h ptr
         declare function edge( byval index as integer ) as shapeEdge ptr
         
         declare sub addVertex( byval v as vec2h ptr )
         declare sub addEdge( byval e as shapeEdge ptr )
         
         declare function transform() as mat3
         declare sub transform( byval M as mat3 )
         
         declare sub render( byval c as uint32 = rgba( 255, 255, 255, 255 ) )
         
         declare sub computeBoundingBox()
      
      private:
         as arrayList   m_vertices '' list of vertices
         as arraylist   m_edges '' list of edges
         as mat3            m_transform = any '' the transformation matrix
         as arrayList   m_boundingBox '' the list of vertices for the bounding box
         as arrayList   m_boundingBoxEdges '' the list of edges for the bounding box
   end class
   
   constructor shape()
      '' load an identity matrix in the transformation matrix when constructed
      m_transform = matrices.identity2h()
   end constructor
   
   destructor shape()
      /'
         this step is necessary because my data structures (like the list)
         don't do garbage collection
      '/
      '' delete vertices
      for i as integer = 0 to m_vertices.count - 1
         dim as vec2h ptr v = m_vertices.get( i )
         
         delete( v )
      next
      
      '' delete edges
      for i as integer = 0 to m_edges.count - 1
         dim as shapeEdge ptr e = m_edges.get( i )
         
         delete( e )
      next i
      
      '' delete bounding box
      for i as integer = 0 to m_boundingBox.count - 1
         dim as vec2h ptr e = m_boundingBox.get( i )
         
         delete( e )
      next i
   end destructor
   
   property shape.vertexCount() as integer
      return( m_vertices.count )
   end property
   
   property shape.edgeCount() as integer
      return( m_edges.count )
   end property
   
   function shape.vertex( byval index as integer ) as vec2h ptr
      return( m_vertices.get( index ) )
   end function
   
   function shape.edge( byval index as integer ) as shapeEdge ptr
      return( m_edges.get( index ) )
   end function
   
   sub shape.addVertex( byval v as vec2h ptr )
      m_vertices.add( v )
   end sub
   
   sub shape.addEdge( byval e as shapeEdge ptr )
      m_edges.add( e )
   end sub
   
   function shape.transform() as mat3
      return( m_transform )
   end function
   
   sub shape.transform( byval M as mat3 )
      m_transform = M
   end sub
   
   sub shape.computeBoundingBox()
      '' minimum and maximum positions of the bounding box
      dim as vec2h tl = vec2h( -100000.0, -100000.0 )
      dim as vec2h rb = vec2h( 100000.0, 100000.0 )
      
      for i as integer = 0 to m_edges.count - 1
         '' iterate through every edge
         dim as shapeEdge ptr e = m_edges.get( i )
         
         dim as vec2h v1 = *( e->v1 )
         dim as vec2h v2 = *( e->v2 )
         
         if( v1.x > tl.x ) then tl.x = v1.x
         if( v1.y > tl.y ) then tl.y = v1.y
         
         if( v2.x < rb.x ) then rb.x = v2.x
         if( v2.y < rb.y ) then rb.y = v2.y
      next
      
      '' create the 4 corners of the bounding box...
      m_boundingBox.add( new vec2h( tl.x, tl.y ) )
      m_boundingBox.add( new vec2h( tl.x, rb.y ) )
      m_boundingBox.add( new vec2h( rb.x, rb.y ) )
      m_boundingBox.add( new vec2h( rb.x, tl.y ) )
      
      '' ...and connect them with edges
      m_boundingBoxEdges.add( new shapeEdge( m_boundingBox.get( 0 ), m_boundingBox.get( 1 ) ) )
      m_boundingBoxEdges.add( new shapeEdge( m_boundingBox.get( 1 ), m_boundingBox.get( 2 ) ) )
      m_boundingBoxEdges.add( new shapeEdge( m_boundingBox.get( 2 ), m_boundingBox.get( 3 ) ) )
      m_boundingBoxEdges.add( new shapeEdge( m_boundingBox.get( 3 ), m_boundingBox.get( 0 ) ) )
   end sub
   
   sub shape.render( byval c as uint32 = rgba( 255, 255, 255, 255 ) )
      /'
         render the shape
         
         the process is simple: transform each vertex with the provided
         matrix (via multiplying), and use that transformed vector to
         draw the shape
      '/
      for i as integer = 0 to m_edges.count - 1
         '' transform both edge vectors first
         dim as shapeEdge ptr e = m_edges.get( i )
         
         dim as vec2h v1 = vec2h( m_transform * ( *e->v1 ) )
         dim as vec2h v2 = vec2h( m_transform * ( *e->v2 ) )
                  
         '' and then draw the edge
         line( v1.x, v1.y ) - ( v2.x, v2.y ), c
      next
      
      '' if the bounding box has been computed, draw it
      if( m_boundingBox.count ) > 0 then
         for i as integer = 0 to m_boundingBoxEdges.count - 1
            '' transform both edge vertices first
            dim as shapeEdge ptr e = m_boundingBoxEdges.get( i )
            
            dim as vec2h v1 = vec2h( m_transform * ( *e->v1 ) )
            dim as vec2h v2 = vec2h( m_transform * ( *e->v2 ) )
                     
            '' and then draw the edge
            line( v1.x, v1.y ) - ( v2.x, v2.y ), rgba( 128, 128, 128, 255 )
         next
      end if
   end sub
#endif
