#ifndef __vec2h__
   #define __vec2h__
   /'
      homogeneous 2D vector type
      
      | x |
      | y |
      | 1 |
      
   '/
   #include once "core.bi"
   #include once "math.bi"
   
   type vec2h
      public:
         as double x
         as double y
         as double w
         
         declare constructor()
         declare constructor( byval nx as double, byval ny as double, byval nw as double = 1.0 )
         declare constructor( byref nv as vec2h )
         declare operator let( byref rhs as vec2h )
                  
         declare function dot( byref b as vec2h ) as double
         declare function cross( byref b as vec2h ) as double
         declare function length() as double
         declare function squaredLength() as double
         declare function unit() as vec2h
         declare sub makeHomogeneous()
         declare sub normalize()
         declare sub turnLeft()
         declare sub turnRight()
         declare sub rotate( byval rAngle as double )
         declare function angle() as double
         declare sub setLength( byval l as double )
         
         declare operator cast() as string
   end type
   
   constructor vec2h() export
      x = 0.0
      y = 0.0
      w = 1.0
   end constructor
   
   constructor vec2h( byval nx as double, byval ny as double, byval nw as double = 1.0 ) export
      x = nx
      y = ny
      w = nw
   end constructor
   
   constructor vec2h( byref nv as vec2h ) export
      x = nv.x
      y = nv.y
      w = nv.w
   end constructor
   
   operator vec2h.let( byref rhs as vec2h ) export
      x = rhs.x
      y = rhs.y
      w = rhs.w
   end operator
   
   '' basic arithmetic operators
   operator + ( byref lhs as vec2h, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs.x + rhs.x, lhs.y + rhs.y ) )
   end operator
   
   operator - ( byref lhs as vec2h, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs.x - rhs.x, lhs.y - rhs.y ) )
   end operator
   
   operator - ( byref lhs as vec2h ) as vec2h export
      return( vec2h( -lhs.x, -lhs.y ) )
   end operator
   
   operator * ( byref lhs as vec2h, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs.x * rhs.x, lhs.y * rhs.y ) )
   end operator
   
   operator * ( byref lhs as vec2h, byref rhs as double ) as vec2h export
      return( vec2h( lhs.x * rhs, lhs.y * rhs ) )
   end operator
   
   operator * ( byref lhs as double, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs * rhs.x, lhs * rhs.y ) )
   end operator
   
   operator * ( byref lhs as vec2h, byref rhs as integer ) as vec2h export
      return( vec2h( lhs.x * rhs, lhs.y * rhs ) )
   end operator
   
   operator * ( byref lhs as integer, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs * rhs.x, lhs * rhs.y ) )
   end operator
   
   operator / ( byref lhs as vec2h, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs.x / rhs.x, lhs.y / rhs.y ) )
   end operator
   
   operator / ( byref lhs as vec2h, byref rhs as double ) as vec2h export
      return( vec2h( lhs.x / rhs, lhs.y / rhs ) )
   end operator
   
   operator / ( byref lhs as double, byref rhs as vec2h ) as vec2h export
      return( vec2h( lhs / rhs.x, lhs / rhs.y ) )
   end operator
   
   operator vec2h.cast() as string export
      return(   "[" & str( this.x ) & "]" & chr( 10 ) & chr( 13 ) & _
                  "[" & str( this.y ) & "]" & chr( 10 ) & chr( 13 ) & _
                  "[" & str( w ) & "]" )
   end operator
   
   function vec2h.dot( byref b as vec2h ) as double export
      '' returns the dot product of this vector with another vector b
      return( x * b.x + y * b.y )
   end function
   
   function dot( byref v as vec2h, byref w as vec2h ) as double export
      return( v.x * w.x + v.y * w.y )
   end function
   
   function vec2h.cross( byref b as vec2h ) as double export
      /'
         the cross product is not defined in 2d, so this function returns the z component
         of the cross product of this vector with vector b, augmented to 3d
      '/
      return( x * b.y - y * b.x )
   end function
   
   function vec2h.length() as double export
      '' returns the length of this vector
      return( sqr( x * x + y * y ) )
   end function
   
   function vec2h.squaredLength() as double export
      /'
         returns the squared length of this vector
         useful when one just want to compare two vectors to see which is longest,
         as this avoids computing the square root
      '/
      return( x * x + y * y )
   end function
   
   function vec2h.unit() as vec2h export
      '' return a normalized copy of this vector, without normalizing the vector itself
      return vec2h( x, y ) / sqr( x * x + y * y )
   end function
   
   sub vec2h.normalize() export
      '' normalizes this vector
      dim as double l = sqr( x * x + y * y )
      
      x /= l
      y /= l
   end sub
   
   function normalize( byval v as vec2h ) as vec2h export
      '' for compatibility
      v.normalize()
      
      return( v )
   end function
   
   /'
      same with these two functions
      turnLeft and turnRight rotates this vector 90 degrees to the left and right, while the
      functions left and right return this vector rotated 90 degrees without actually rotating it
   '/
   sub vec2h.turnLeft() export
      this = vec2h( y, -x )
   end sub
   
   sub vec2h.turnRight() export
      this = vec2h( -y, x )
   end sub
   
   function turnLeft( byref v as vec2h ) as vec2h export
      return( vec2h( v.y, -v.x ) )
   end function

   function turnRight( byref v as vec2h ) as vec2h export
      return( vec2h( -v.y, v.x ) )
   end function
   
   sub vec2h.rotate( byval rAngle as double ) export
      /'
         rotates this vector by rAngle radians
      '/
      
      dim as double si = sin( rAngle )
      dim as double co = cos( rAngle )
      
      this = vec2h( x * co - y * si, x * si + y * co  )
   end sub
   
   function rotate( byref v as vec2h, byval rAngle as double ) as vec2h export
      /'
         returns the vector v, rotated by rAngle radians
      '/

      dim as double si = sin( rAngle )
      dim as double co = cos( rAngle )
      
      return( vec2h( v.x * co - v.y * si, v.x * si + v.y * co ) )          
   end function
   
   function vec2h.angle() as double export
      '' returns the angle that this vector points to, in radians
      
      return( atan2( y, x ) )
   end function
   
   function vectorAngle overload( byval v as vec2h ) as double export
      '' returns the angle that v vector points to, in radians
      
      return atan2( v.y, v.x )
   end function
   
   function vectorAngle( byval v as vec2h, byval w as vec2h ) as double export
      '' returns the angle between two vectors v and w
      
      dim as double cosineOfAngle = dot( normalize( v ), normalize( w ) )
      
      return( -acos( clamp( -1.0, 1.0, cosineOfAngle ) ) )
   end function
   
   sub vec2h.makeHomogeneous() export
      '' homogeneize this vector (make the w component 1.0)
      
      x /= w
      y /= w
      w /= w
   end sub

   sub vec2h.setLength( byval l as double )
      '' sets the length of this vector
      dim va as double = atan2( y, x )
      
      x = l * cos( va )
      y = l * sin( va )
   end sub
   
   function distance( byval a as vec2h, byval b as vec2h ) as double
      '' returns the distance between two vectors
      return( ( a - b ).length )
   end function
   
   function alignment( byval position as vec2h, byval size as vec2h, byval align as vec2h ) as vec2h
      '' returns a vec2h aligned between position and position + size
      '' align should be a normalized vector
      return( position + size * align )
   end function
   
   function interpolate( byval p0 as vec2h, byval p1 as vec2h, byval t as double ) as vec2h
      '' returns the interpolation point t (0..1) between p0 and p1
      return( ( 1 - t ) * p0 + t * p1 )
   end function
#endif
